@@ -9,6 +9,8 @@ For hosts that do get parsed, their values will be normalized to lowercase ASCII
 
 Note: The suffix reference data is baked into this crate at build time. This reduces the runtime overhead of parsing all that data out, but can also cause implementing apps to grow stale if they haven't been (re)packaged in a while.
 
+
+
 ## Examples
 
 Initiate a new instance using [`Domain::parse`]. If that works, you then have accesses to the individual components:
@@ -28,12 +30,18 @@ assert_eq!(dom.root(), "mydomain");
 assert_eq!(dom.suffix(), "com");
 assert_eq!(dom.tld(), "mydomain.com");
 
-// If you just want the sanitized host back as an owned value, use
-// `Domain::take` or `Domain::into_string`.
-let owned = dom.into_string(); // "www.mydomain.com"
+// If you just want the sanitized host back as an owned value, use `Domain::take`:
+let owned = dom.take(); // "www.mydomain.com"
 ```
 
 A [`Domain`] object can be dereferenced to a string slice representing the sanitized host. You can also consume the object into an owned string with [`Domain::take`].
+
+
+
+## Optional Crate Features
+
+* `serde`: Enables serialization/deserialization support.
+
 */
 
 #![warn(clippy::filetype_is_file)]
@@ -68,10 +76,6 @@ use self::list::{
 	PSL_MAIN,
 	PSL_WILD,
 };
-use smartstring::{
-	LazyCompact,
-	SmartString,
-};
 use std::{
 	cmp::Ordering,
 	convert::TryFrom,
@@ -92,6 +96,7 @@ use std::{
 
 
 
+#[doc(hidden)]
 /// # (Not) Random State.
 ///
 /// Using a fixed seed value for `AHashSet`/`AHashMap` drops a few dependencies
@@ -99,7 +104,7 @@ use std::{
 /// static that would be used otherwise.
 ///
 /// For our purposes, the variability of truly random keys isn't really needed.
-pub(crate) const AHASH_STATE: ahash::RandomState = ahash::RandomState::with_seeds(13, 19, 23, 71);
+pub const AHASH_STATE: ahash::RandomState = ahash::RandomState::with_seeds(13, 19, 23, 71);
 
 
 
@@ -148,11 +153,11 @@ macro_rules! impl_try {
 /// assert_eq!(dom.tld(), "mydomain.com");
 ///
 /// // If you just want the sanitized host back as an owned value, use
-/// // `Domain::take` or `Domain::into_string`.
-/// let owned = dom.into_string(); // "www.mydomain.com"
+/// // `Domain::take`:
+/// let owned = dom.take(); // "www.mydomain.com"
 /// ```
 pub struct Domain {
-	host: SmartString<LazyCompact>,
+	host: String,
 	root: Range<usize>,
 	suffix: Range<usize>,
 }
@@ -241,41 +246,59 @@ impl Domain {
 	///
 	/// Hosts with unknown or missing suffixes are rejected. Otherwise all
 	/// values are normalized to lowercase ASCII.
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// // A regular ASCII domain:
+	/// let dom = Domain::parse("www.MyDomain.com").unwrap();
+	/// assert_eq!(dom.as_str(), "www.mydomain.com");
+	///
+	/// // Non-ASCII domains are normalized to Punycode for consistency:
+	/// let dom = Domain::parse("www.♥.com").unwrap();
+	/// assert_eq!(dom.as_str(), "www.xn--g6h.com");
+	///
+	/// // An incorrectly structured "host" won't parse:
+	/// assert!(Domain::parse("not.a.domain.123").is_none());
+	/// ```
 	pub fn parse<S>(src: S) -> Option<Self>
 	where S: AsRef<str> {
 		idna::domain_to_ascii_strict(src.as_ref().trim_matches(|c: char| c == '.' || c.is_ascii_whitespace()))
 			.ok()
-			.and_then(|host| parse_suffix(&host)
-				.map(|s|
-					// This has a subdomain, i.e. the root is in the middle.
-					if let Some(d) = host.as_bytes()
-						.iter()
-						.take(s - 1)
-						.rposition(|x| x == &b'.')
-					{
-						Self {
-							root: d + 1..s - 1,
-							suffix: s..host.len(),
-							host: host.into(),
-						}
+			.and_then(|host| find_dots(&host)
+				.map(|(mut d, s)| {
+					if 0 < d { d += 1; }
+					Self {
+						root: d..s - 1,
+						suffix: s..host.len(),
+						host,
 					}
-					// The root starts at zero.
-					else {
-						Self {
-							root: 0..s - 1,
-							suffix: s..host.len(),
-							host: host.into(),
-						}
-					}
-				)
+				})
 			)
 	}
+}
 
+/// ## WWW.
+impl Domain {
 	#[must_use]
 	/// # Has Leading WWW.
 	///
 	/// This will return `true` if the domain begins with "www." _and_ that
 	/// "www." is a subdomain. (The latter is usually but not always the case!)
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom1 = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert!(dom1.has_www());
+	///
+	/// let dom2 = Domain::parse("blobfolio.com").unwrap();
+	/// assert!(! dom2.has_www());
+	/// ```
 	pub fn has_www(&self) -> bool {
 		self.root.start >= 4 && self.host.starts_with("www.")
 	}
@@ -285,8 +308,24 @@ impl Domain {
 	///
 	/// This will return a clone of the instance without the leading WWW if it
 	/// has one, otherwise `None`.
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom1 = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert_eq!(dom1.as_str(), "www.blobfolio.com");
+	///
+	/// let dom2 = dom1.without_www().unwrap();
+	/// assert_eq!(dom2.as_str(), "blobfolio.com");
+	/// ```
 	pub fn without_www(&self) -> Option<Self> {
 		if self.has_www() {
+			// Manual assignment via clone/split works around a strange issue
+			// affecting string slice indexing of the host value in cargo-test
+			// in 1.53.0. The problem doesn't trigger in nightly, so we might
+			// revert after the next release.
 			Some(Self {
 				host: self.host[4..].into(),
 				root: self.root.start - 4..self.root.end - 4,
@@ -299,22 +338,12 @@ impl Domain {
 
 /// # Conversion.
 impl Domain {
-	#[must_use]
-	/// # Into String.
-	///
-	/// Consume the struct, returning the sanitized host as an owned string.
-	pub fn into_string(self) -> String { self.host.into() }
-
 	#[allow(clippy::missing_const_for_fn)] // Doesn't work.
 	#[must_use]
-	/// # Into `SmartString`.
+	/// # Take String
 	///
-	/// Consume the struct, returning the sanitized host as an owned
-	/// `SmartString` (which is how this struct stores the data).
-	///
-	/// If you would rather have a regular `String`, use [`Domain::into_string`]
-	/// instead.
-	pub fn take(self) -> SmartString<LazyCompact> { self.host }
+	/// Consume the struct, returning the sanitized host as an owned `String`.
+	pub fn take(self) -> String { self.host }
 }
 
 /// # Getters.
@@ -324,6 +353,15 @@ impl Domain {
 	///
 	/// Return the sanitized host as a string slice. This is equivalent to
 	/// dereferencing the object.
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert_eq!(dom.host(), "www.blobfolio.com");
+	/// ```
 	pub fn host(&self) -> &str { &self.host }
 
 	#[must_use]
@@ -331,6 +369,15 @@ impl Domain {
 	///
 	/// Return the root portion of the host, if any. This does not include any
 	/// leading or trailing periods.
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert_eq!(dom.root(), "blobfolio");
+	/// ```
 	pub fn root(&self) -> &str {
 		&self.host[self.root.start..self.root.end]
 	}
@@ -340,6 +387,15 @@ impl Domain {
 	///
 	/// Return the subdomain portion of the host, if any. This does not include
 	/// any trailing periods.
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert_eq!(dom.subdomain(), Some("www"));
+	/// ```
 	pub fn subdomain(&self) -> Option<&str> {
 		if self.root.start > 0 { Some(&self.host[0..self.root.start - 1]) }
 		else { None }
@@ -350,6 +406,15 @@ impl Domain {
 	///
 	/// Return the suffix of the host. This does not include any leading
 	/// periods.
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert_eq!(dom.suffix(), "com");
+	/// ```
 	pub fn suffix(&self) -> &str {
 		&self.host[self.suffix.start..self.suffix.end]
 	}
@@ -359,6 +424,15 @@ impl Domain {
 	///
 	/// Return the TLD portion of the host, i.e. everything but the
 	/// subdomain(s).
+	///
+	/// ## Examples
+	///
+	/// ```
+	/// use adbyss_psl::Domain;
+	///
+	/// let dom = Domain::parse("www.blobfolio.com").unwrap();
+	/// assert_eq!(dom.tld(), "blobfolio.com");
+	/// ```
 	pub fn tld(&self) -> &str {
 		&self.host[self.root.start..]
 	}
@@ -366,7 +440,32 @@ impl Domain {
 
 
 
-/// # Find Suffix.
+#[cfg(any(test, feature = "serde"))]
+impl serde::Serialize for Domain {
+	/// # Serialize.
+	///
+	/// Use the optional `serde` crate feature to enable serialization support.
+	fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+	where S: serde::Serializer {
+		self.host.serialize(serializer)
+	}
+}
+
+#[cfg(any(test, feature = "serde"))]
+impl<'de> serde::Deserialize<'de> for Domain {
+	/// # Deserialize.
+	///
+	/// Use the optional `serde` crate feature to enable serialization support.
+	fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+	where D: serde::de::Deserializer<'de> {
+		let s: &str = serde::de::Deserialize::deserialize(deserializer)?;
+		Self::parse(s).ok_or_else(|| serde::de::Error::custom("Invalid domain."))
+	}
+}
+
+
+
+/// # Find Dots.
 ///
 /// The hardest part of suffix validation is teasing the suffix out of the
 /// hostname. Odd.
@@ -374,39 +473,42 @@ impl Domain {
 /// The suffix cannot be the whole of the thing, but should be the biggest
 /// matching chunk of the host.
 ///
-/// If a match is found, the starting index of the suffix (after its dot) is
-/// returned.
-fn parse_suffix(host: &str) -> Option<usize> {
+/// If a match is found, the location of the start of the root (its dot, or zero)
+/// is returned along with the starting index of the suffix (after its dot).
+fn find_dots(host: &str) -> Option<(usize, usize)> {
 	let len: usize = host.len();
 	if len < 3 || PSL_WILD.contains_key(host) || PSL_MAIN.contains(host) { return None; }
 
-	let bytes: &[u8] = host.as_bytes();
+	let mut last: usize = 0;
 	let mut dot: usize = 0;
-	for idx in 0..len {
-		if bytes[idx] != b'.' { continue; }
+	for (idx, _) in host.as_bytes().iter().enumerate().filter(|(_, &b)| b'.' == b) {
+		// We know there is at least one more byte.
+		let rest: &str = unsafe { host.get_unchecked(idx + 1..) };
 
 		// This is a wild extension.
-		if let Some(exceptions) = PSL_WILD.get(&host[idx + 1..]) {
+		if let Some(exceptions) = PSL_WILD.get(rest) {
 			// Our last chunk might start at zero instead of dot-plus-one.
 			let after_dot: usize =
 				if dot == 0 { 0 }
 				else { dot + 1 };
 
 			// This matches an exception, making the found suffix the true
-			// suffix.
-			if exceptions.contains(&&host[after_dot..idx]) {
-				return Some(idx + 1);
+			// suffix. Note: there cannot be a dot at position zero, so the
+			// range is always valid.
+			if exceptions.contains(&unsafe { host.get_unchecked(after_dot..idx) }) {
+				return Some((dot, idx + 1));
 			}
 			// There has to be a before-before part.
 			else if dot == 0 { return None; }
 			// Otherwise the last chunk is part of the suffix.
-			return Some(after_dot);
+			return Some((last, after_dot));
 		}
 		// This is a normal suffix.
-		else if PSL_MAIN.contains(&host[idx + 1..]) {
-			return Some(idx + 1);
+		else if PSL_MAIN.contains(rest) {
+			return Some((dot, idx + 1));
 		}
 
+		std::mem::swap(&mut dot, &mut last);
 		dot = idx;
 	}
 
@@ -418,6 +520,7 @@ fn parse_suffix(host: &str) -> Option<usize> {
 #[cfg(test)]
 mod tests {
 	use super::*;
+	use brunch as _;
 
 	#[test]
 	/// # Test TLD Parsing.
@@ -541,15 +644,16 @@ mod tests {
 		assert_eq!(dom.root(), "xn--85x722f");
 		assert_eq!(dom.suffix(), "xn--fiqs8s");
 		assert_eq!(dom.tld(), "xn--85x722f.xn--fiqs8s");
-		assert_eq!(dom.deref(), "abc.www.xn--85x722f.xn--fiqs8s");
 		assert_eq!(dom.host(), "abc.www.xn--85x722f.xn--fiqs8s");
 
+		// Make sure dereference does the right thing. It should...
+		assert_eq!(dom.host(), dom.deref());
+
 		dom = Domain::parse("blobfolio.com").unwrap();
 		assert_eq!(dom.subdomain(), None);
 		assert_eq!(dom.root(), "blobfolio");
 		assert_eq!(dom.suffix(), "com");
 		assert_eq!(dom.tld(), "blobfolio.com");
-		assert_eq!(dom.deref(), "blobfolio.com");
 		assert_eq!(dom.host(), "blobfolio.com");
 
 		dom = Domain::parse("www.blobfolio.com").unwrap();
@@ -557,16 +661,53 @@ mod tests {
 		assert_eq!(dom.root(), "blobfolio");
 		assert_eq!(dom.suffix(), "com");
 		assert_eq!(dom.tld(), "blobfolio.com");
-		assert_eq!(dom.deref(), "www.blobfolio.com");
 		assert_eq!(dom.host(), "www.blobfolio.com");
 
-		assert!(dom.has_www());
-		dom = dom.without_www().expect("Dom without www.");
+		// Test a long subdomain.
+		dom = Domain::parse("another.damn.sub.domain.blobfolio.com").unwrap();
+		assert_eq!(dom.subdomain(), Some("another.damn.sub.domain"));
+		assert_eq!(dom.root(), "blobfolio");
+		assert_eq!(dom.suffix(), "com");
+		assert_eq!(dom.tld(), "blobfolio.com");
+		assert_eq!(dom.host(), "another.damn.sub.domain.blobfolio.com");
+
+		// Also make sure stripping works OK.
+		dom = Domain::parse("    ....blobfolio.com....    ").unwrap();
 		assert_eq!(dom.subdomain(), None);
 		assert_eq!(dom.root(), "blobfolio");
 		assert_eq!(dom.suffix(), "com");
 		assert_eq!(dom.tld(), "blobfolio.com");
-		assert_eq!(dom.deref(), "blobfolio.com");
 		assert_eq!(dom.host(), "blobfolio.com");
 	}
+
+	#[test]
+	/// # Test WWW Stripping.
+	fn t_without_www() {
+		let dom1 = Domain::parse("www.blobfolio.com").unwrap();
+		assert!(dom1.has_www());
+
+		let dom2 = dom1.without_www().unwrap();
+		assert_eq!(dom2.subdomain(), None);
+		assert_eq!(dom2.root(), "blobfolio");
+		assert_eq!(dom2.suffix(), "com");
+		assert_eq!(dom2.tld(), "blobfolio.com");
+		assert_eq!(dom2.host(), "blobfolio.com");
+		assert!(! dom2.has_www());
+	}
+
+	#[test]
+	/// # Serde tests.
+	fn t_serde() {
+		let dom1: Domain = Domain::parse("serialize.domain.com")
+			.expect("Domain failed.");
+
+		// Serialize it.
+		let serial: String = serde_json::to_string(&dom1)
+			.expect("Serialize failed.");
+		assert_eq!(serial, "\"serialize.domain.com\"");
+
+		// Deserialize it.
+		let dom2: Domain = serde_json::from_str(&serial).expect("Deserialize failed.");
+		assert_eq!(dom1, dom2);
+	}
 }
